/*****************************************************************************
 * NAME: bbslashdot
 * DESCRIPTION: displays slashdot articles on the screen
 * NEEDS: XLib, G++ 3.x and libxml2
 *
 * (C) 2003 by Ruediger Sonderfeld <cplusplushelp@gmx.net>
 *
 * bbslashdot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bbslashdot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with bbslashdot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************/

#include "config.h"

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <memory>
#include <sstream>
extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <libxml/parser.h>
#include <libxml/xmlmemory.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h> //no csignal because we need the POSIX stuff
}

#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#ifndef WGET_PATH
#ifndef NO_WGET
#define WGET_PATH "/usr/bin/wget"
#endif
#endif

typedef RETSIGTYPE (*handler_t)(int);

class error
{
  std::string str;
public:
  inline error(const std::string &err)
    : str(err) { }
  inline error(const error &obj)
    : str(obj.str) { }
  virtual ~error() { }
  error &operator=(const error &obj)
  {
    if(this!=&obj)
      str=obj.str;
    return *this;
  }
  virtual inline const char *what(void) const { return "error"; }
  virtual inline const std::string &description(void) const { return str; }
};

class xmlerror : public error
{
  std::string str;
public:
  inline xmlerror(const std::string &err)
    : error(err) { }
  ~xmlerror() { }
  inline xmlerror(const error &obj)
    : error(obj) { }
  xmlerror &operator=(const xmlerror &obj)
  {
    error::operator=(obj);
    return *this;
  }
};

class window
{
  Window win_;
  GC gc;
  Display *display_;
  int x_, y_;

  window &operator=(const window&) { return *this; }
public:
  window(const window &obj) 
    : win_(obj.win_), gc(obj.gc), display_(obj.display_), x_(obj.x_), 
      y_(obj.y_)
  { }
  window(Display *display, int screen, Window parent, int x, int y, int width,
	 int height, unsigned int border_width, const XColor &border, 
	 const XColor &bg);
  ~window() 
  { 
    XSelectInput(display_,win_,None); 
    XDestroyWindow(display_,win_); 
  }
  void map(void)
  {
    XMapWindow(display_,win_);
    XFlush(display_);
  }
  void winclear(void)
  {
    XClearWindow(display_,win_);
    XFlush(display_);
  }

protected:
  inline Window getWindow(void) const{ return win_; }
  inline GC getGC(void) const { return gc; }
  inline Display *getDisplay(void) const { return display_; }
  inline int x(void) const { return x_; }
  inline int y(void) const { return y_; }
};

window::window(Display *display, int screen, Window parent, int x, int y, 
	       int width, int height, unsigned int border_width, const XColor &
	       border,const XColor &background)
  : display_(display), x_(x), y_(y)
{
  XSetWindowAttributes attribute;
  attribute.border_pixel=border.pixel;
  attribute.background_pixel=background.pixel;
#ifdef USE_SINGLE_WINDOW
  attribute.override_redirect=False;
#else
  attribute.override_redirect=True;
#endif

  win_=XCreateWindow(display,parent,x,y,width,height,
		     border_width,DefaultDepth(display,screen),
		     InputOutput,
		     CopyFromParent,
		     CWBackPixel|CWBorderPixel|CWOverrideRedirect,
		     &attribute);
  XSizeHints shints;
  shints.flags=PPosition | PSize;
  shints.x=x;
  shints.y=y;
  shints.width=width;
  shints.height=height;
  XSetNormalHints(display,win_,&shints);

  XWMHints whints;
  whints.flags=InputHint;
  whints.input=True;
  XSetWMHints(display_,win_,&whints);

  XSelectInput(display_,win_,ExposureMask | KeyPressMask | ButtonPressMask );
  XCreateFontCursor(display_, XC_arrow);
  XStoreName(display_,win_,"xslashdot");
  gc=XCreateGC(display_,win_,0,NULL);
}

template<class charT, class traits=std::char_traits<charT> >
class windowstreambuf : public std::basic_streambuf<charT,traits>, 
			private window
{
  static const size_t BUFFER_SIZE=100;
  char buffer[BUFFER_SIZE];
  size_t pos;
  int x_, y_;
  int init_x, init_y;
  XFontStruct *font;
public:
  windowstreambuf(window *win, const char *font_p, const XColor &text)
    : window(*win), pos(0), x_(0), y_(15), init_x(0), init_y(15),
      font(XLoadQueryFont(getDisplay(),font_p))
  {
    XSetFont(getDisplay(),getGC(),font->fid);
    XSetForeground(getDisplay(),getGC(),text.pixel);
  }
  windowstreambuf(window *win, const char *font_p, const XColor &text,
		  int x, int y)
    : window(*win), pos(0), x_(x), y_(y), init_x(x), init_y(y), 
      font(XLoadQueryFont(getDisplay(),font_p))
  {
    XSetFont(getDisplay(),getGC(),font->fid);
    XSetForeground(getDisplay(),getGC(),text.pixel);
  }
  ~windowstreambuf()
  { XUnloadFont(getDisplay(),font->fid); }

  inline void setx(int x) { x_=x; }
  inline void sety(int y) { y_=y; }
  inline void rewind(void) { x_=init_x; y_=init_y; }
  inline void clear(void) { rewind(); winclear(); }
protected:
  int_type overflow(int_type c=traits::eof())
  {
    if(pos>=BUFFER_SIZE-1)
      sync();
    if(c=='\n')
    {
      sync();
      y_+=15; //replace 15 with font height
      x_=init_x;
      return c;
    }
    buffer[pos++]=c;
    return c;
  }
  int sync(void)
  {
    if(pos==0)
      return true;
    XDrawString(getDisplay(),getWindow(),getGC(),x_,y_,buffer,pos);
    XFlush(getDisplay());
    x_+=XTextWidth(font,buffer,pos);
    pos=0;
    return true;
  }
};

struct manipulator
{
  enum manipulator_t
    {
      REWIND, CLEAR
      //, CHANGEFONT, SETX, SETY ...
    } manipulator_m;
  manipulator(enum manipulator_t v)
    : manipulator_m(v) { }
};

manipulator xrewind(manipulator::REWIND);
manipulator xclear(manipulator::CLEAR);

std::ostream &operator<<(std::ostream &out, const manipulator &obj)
{
  windowstreambuf<char> *buf=dynamic_cast<windowstreambuf<char>*>(out.rdbuf());
  if(buf!=NULL)
  {
    switch(obj.manipulator_m)
    {
    case manipulator::REWIND:
      buf->rewind();
      break;
    case manipulator::CLEAR:
      buf->clear();
    }
  }
  return out;
}

#ifndef NO_WGET
void http_get(const char *host, const std::string &to)
{
  switch(vfork())
  {
  case -1:
    throw error(strerror(errno));
  case 0:
    if(execl(WGET_PATH,WGET_PATH,host,"-O",to.c_str(),NULL)==-1)
      throw error(strerror(errno));
  default:
    {
      int status;
      wait(&status);
      if(!WIFEXITED(status))
	throw error("wget failed");
    }
  }
}

//find a good temporary filename
std::string mktmpname(const std::string &prefix)
{
#ifndef NO_SECURE_TMP
  std::string buffer;
  for(;;)
  {
    std::stringstream postfix;
    postfix << (10000.0*rand()/(RAND_MAX+1.0)*getpid()/getuid()*
		(getpid()*rand()/(RAND_MAX+1.0)));
    buffer=prefix+"-"+postfix.str()+".tmp";
    struct stat st;
    if(lstat(buffer.c_str(),&st)==-1)
      if(errno==ENOENT)
	break;
  }
  return buffer;
#else
  return prefix+".tmp";
#endif //NO_SECURE_TMP
}
#endif //NO_WGET

class slashdot
{
  std::ostream &out;
  const char *file_;
  xmlDocPtr doc;
  xmlNodePtr node, start;

  int refresh_;
  int until_next;
  long timer;
  int next;

#ifndef NO_WGET
  static const char * const tmpfile_prefix;
  std::string tmpfile_;
#endif
  static slashdot *this_; // I know this sucks! But there should only be 1 
                          //instance from slashdot
public:
  static const char *const SLASHDOT_XML;

  // read slashdot.xml from File
  slashdot(std::ostream &out_p, long narticletimer, const char *file, 
	   int refresh=30)
    : out(out_p),file_(file),doc(xmlParseFile(file)),refresh_(refresh), 
      timer(narticletimer),next(refresh_/(narticletimer/60))
  {
    if(!doc)
      throw xmlerror("couldn't open file");
    init_xml();
  }
#ifndef NO_WGET
  // GET slashdot.xml via HTTP
  slashdot(std::ostream &out_p, long narticletimer, int refresh=30)
    : out(out_p),file_(NULL),refresh_(refresh),timer(narticletimer),
      next( refresh_/(narticletimer/60) ), tmpfile_(mktmpname(tmpfile_prefix))
  {
    if(refresh<30)
      throw error("slashdot doesn't allow to update the slashdot.xml file more"
		  " than once every 30 minutes");
    http_get(SLASHDOT_XML,tmpfile_);
    if(!(doc=xmlParseFile(tmpfile_.c_str())))
      throw xmlerror("couldn't open file");
    init_xml();
  }
#endif //NO_WGET
  inline ~slashdot() { }
  void print_next_article(void)
  {
    node=node->next;
    reprint_article();
  }
  void print_prev_article(void)
  {
    if( (node=node->prev) !=NULL)
      if( (node=node->prev) ==NULL)
	node=start;
    reprint_article();
  }
  void reprint_article(void);
  void openurl(const char *browser_string);

  inline static handler_t handler(slashdot *this_p)
  { 
    this_=this_p;
    return this_->nexthandler;
  }

private:
  void init_xml(void)
  {
    if( (node=xmlDocGetRootElement(doc)) ==NULL)
      throw xmlerror("couldn't get root element");
    if( (start=node=node->children) ==NULL)
      throw xmlerror("wrong format");
  }

  static RETSIGTYPE nexthandler(int)
  {
    this_->print_next_article();
    if(this_->until_next==this_->next)
    {
      this_->reloadxml();
      this_->until_next=0;
    }
    ++this_->until_next;
#ifdef USE_OLD_TIMER
    alarm(static_cast<unsigned int>(this_->timer));
#endif
  }

  void reloadxml(void)
  {
#ifndef NO_WGET
    if(file_)
    {
#endif
      if(!(doc=xmlParseFile(file_)))
	throw xmlerror("couldn't open file");
#ifndef NO_WGET
    }
    else
    {
      http_get(SLASHDOT_XML,tmpfile_);
      if(!(doc=xmlParseFile(tmpfile_.c_str())))
	throw xmlerror("couldn't open file");
    }
#endif
    init_xml();
  }
};

const char *const slashdot::tmpfile_prefix="/tmp/slashdot.xml";
const char *const slashdot::SLASHDOT_XML="http://slashdot.org/slashdot.xml";
slashdot *slashdot::this_=NULL;

void slashdot::openurl(const char *browser_string)
{
  //get URL
  char *url;
  for(;node!=NULL &&
	xmlStrcmp(node->name,reinterpret_cast<const xmlChar*>("story"));)
    node=node->next;
  
  xmlNodePtr snode=node->children;
  while(snode!=NULL)
  {
    if(!xmlStrcmp(snode->name,(const xmlChar*)"url"))
      url=reinterpret_cast<char*>(xmlNodeListGetString(doc,
						       snode->xmlChildrenNode, 
						       1));
    snode=snode->next;
  }

  if(url==NULL)
    throw error("internal error");

  //start Browser
  char *browser;
  if(strstr(browser_string,"%s")==NULL)
  {
    browser=new char[strlen(browser_string)+strlen(url)+2];
    sprintf(browser,"%s %s",browser_string,url);
  }
  else
  {
    browser=new char[strlen(browser_string)+strlen(url)];
    sprintf(browser,browser_string,url);
  }

  char *array[6];
  if( (array[0]=strtok(browser," ")) ==NULL)
    throw error("wrong format");

  size_t i=1;
  for(;i<5&&(array[i]=strtok(NULL," "));++i) ;
  array[i]=NULL;

  switch(fork())
  {
  case -1:
    throw error(strerror(errno));
  case 0:
    if(execvp(array[0],array))
      throw error(strerror(errno));
  default:
    {
      int status;
      wait(&status);
      if(!WIFEXITED(status))
        throw error("browser failed");
    }
  }

  delete[] browser;
}

void slashdot::reprint_article(void)
{
  for(bool restarted=false;node!=NULL &&
	xmlStrcmp(node->name,reinterpret_cast<const xmlChar*>("story"));)
  {
    if( (node=node->next) ==NULL )
      if(!restarted)
      {
	restarted=true;
	node=start;
      }
      else
	throw error("internal error");
  }

  out << xclear;
  xmlNodePtr snode=node->children;
  while(snode!=NULL)
  {
    if(!xmlStrcmp(snode->name,reinterpret_cast<const xmlChar*>("title")))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "title: " << key << '\n';
      xmlFree(key);
    }
    else if(!xmlStrcmp(snode->name,(const xmlChar*)"url"))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "url: " << key << '\n';
      xmlFree(key);
    }
    else if(!xmlStrcmp(snode->name,(const xmlChar*)"time"))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "time: " << key << '\n';
      xmlFree(key);
    }
    else if(!xmlStrcmp(snode->name,(const xmlChar*)"author"))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "author: " << key << '\n';
      xmlFree(key);
    }
    else if(!xmlStrcmp(snode->name,(const xmlChar*)"department"))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "department: " << key << '\n';
      xmlFree(key);
    }
    else if(!xmlStrcmp(snode->name,(const xmlChar*)"section"))
    {
      xmlChar *key=xmlNodeListGetString(doc,snode->xmlChildrenNode, 1);
      out << "section: " << key << '\n';
      xmlFree(key);
    }
    snode=snode->next;
  }
}

void getcolor(Display *display,int screen,const std::string &in, 
	      XColor &out)
{
  Colormap cmap=XDefaultColormap(display,screen);
  if(in[0]=='#')
  {
    out.flags=DoRed | DoGreen | DoBlue;
    //TODO: replace this construct:
    std::stringstream red(in.substr(1,2));
    red >> out.red;
    std::stringstream green(in.substr(3,2));
    green >> out.green;
    std::stringstream blue(in.substr(5,2));
    blue >> out.blue;
    if(!XAllocColor(display,cmap,&out))
      throw error("coudln't allocate color");
  }
  else
  {
    XColor exact;
    if(!XAllocNamedColor(display,cmap,in.c_str(),&out,&exact))
      throw error("coudln't allocate color");
    if(exact.red!=out.red || exact.green!=out.green || exact.blue!=out.blue)
      std::cerr << "warning: color not exact" << std::endl;
  }
}

void configfile(Display *display, int screen, const char *filename, 
		int &x, int &y, int &width, int &height,long &alarm_s, 
		int &refresh, std::string &file,std::string &browser, 
		XColor &border_color, XColor &win_color,XColor &text_color)
{
  std::ifstream in(filename);
  if(!in.is_open())
    return;
  for(;;)
  {
    std::string line;
    if(!std::getline(in,line))
      break;
    if(line[0]=='#')
      continue;
    std::string right;
    std::string left;
    std::string *out=&left;
    for(std::string::const_iterator i=line.begin();i!=line.end();++i)
    {
      switch(*i)
      {
      case ' ':
      case '\t':
	continue;
      case '!':
	break;
      case ':':
	if(out==&right)
	  return;
	out=&right;
	continue;
      }
      *out+=*i;
    }
    if(left=="xslashdot.window.x")
      x=atoi(right.c_str());
    else if(left=="xslashdot.window.y")
      y=atoi(right.c_str());
    else if(left=="xslashdot.window.width")
      width=atoi(right.c_str());
    else if(left=="xslashdot.window.height")
      height=atoi(right.c_str());
    else if(left=="xslashdot.window.color.border")
      getcolor(display,screen,right,border_color);
    else if(left=="xslashdot.window.color.text")
      getcolor(display,screen,right,text_color);
    else if(left=="xslashdot.window.color.window")
      getcolor(display,screen,right,win_color);
    else if(left=="xslashdot.next")
      alarm_s=atol(right.c_str());
    else if(left=="xslashdot.browser")
      browser=right;
    else if(left=="xslashdot.file")
      file=right;
    else if(left=="xslashdot.refresh")
      refresh=atoi(right.c_str());
    else
      throw error("unkown configure option");
  }
}

static const size_t BUFFER_SIZE=100;

int main(int argc, char **argv)
{
  try
  {
    int x=0,y=0,width=400,height=100;
    std::string file;
    long alarm_s=60;
    std::string browser="mozilla %s";//-remote 'openURL(%s)'";
    std::string config;
    char *home=getenv("HOME");
    if(!home)
      config=".xslashdotrc";
    else
      config=std::string(home)+"/.xslashdotrc";
    bool use_config=true, use_file=true;
    
    if(argc>1)
    {
      int i=1;
      for(;i<argc;++i)
      {
	if(argv[i][0]=='-')
	{
	  if(argv[i][1]=='x')
	    x=atoi(argv[++i]);
	  else if(argv[i][1]=='y')
	    y=atoi(argv[++i]);
	  else if(argv[i][1]=='p')
	    use_config=false;
	  else if(argv[i][1]=='d')
	    use_file=false;
	  else if(argv[i][1]=='n')
	    alarm_s=atol(argv[++i]);
	  else if(argv[i][1]=='w'||!strcmp(argv[i],"--width"))
	    width=atoi(argv[++i]);
	  else if(argv[i][1]=='h'||!strcmp(argv[i],"--height"))
	    height=atoi(argv[++i]);
	  else if(argv[i][1]=='f'||!strcmp(argv[i],"--file"))
	    file=argv[++i];
	  else if(argv[i][1]=='b'||!strcmp(argv[i],"--browser"))
	    browser=argv[++i];
	  else if(argv[i][1]=='c'||!strcmp(argv[i],"--config"))
	    config=argv[++i];
	  else if(!strcmp(argv[i],"--help"))
	  {
	    std::cout << "usage: " << argv[0] 
		      << "-x x -y y [-h|--height] height [-w|--width] width "
	      "[-f|--file] slashdot-xml-file [-n] sec [-b|--browser] browser "
	      "[-c|--config] configfile [-pd]"
		      << std::endl;
	  }
	  else
	  {
	    std::cerr << "wrong argument: " << argv[i] << "\nsee " << argv[0]
		      << " --help" << std::endl;
	    return 1;
	  }
	}
	else
	{
	  std::cerr << "wrong argument: " << argv[i] << "\nsee " << argv[0] << 
	    "--help" << std::endl;
	  return 1;
	}
      }
    }

    Display *display=XOpenDisplay(NULL);
    if(display==NULL)
      return 1;
    int screen=DefaultScreen(display);

    int refresh=30;
    XColor border_c,window_c,text_c;

    border_c.pixel=WhitePixel(display,screen);
    window_c.pixel=BlackPixel(display,screen);
    text_c.pixel=WhitePixel(display,screen);

    if(use_config)
      configfile(display,screen,config.c_str(),x,y,width,height,alarm_s,refresh
		 ,file,browser,border_c,window_c,text_c);

#ifdef NO_WGET
    if(file.empty())
      throw error("you need to set the file to parse or enable wget support");
#endif
    window win(display,screen,RootWindow(display,screen),x,y,width,height,1,
	       border_c,window_c);
    win.map();

    windowstreambuf<char> buf(&win,"6x10",text_c,5,15);
    std::ostream out(&buf);

    std::auto_ptr<slashdot> sdconn;

#ifndef NO_WGET
    if(!file.empty()&&use_file)
#endif
      sdconn=std::auto_ptr<slashdot>(new slashdot(out,alarm_s,file.c_str(),
						  refresh));
#ifndef NO_WGET
    else
      sdconn=std::auto_ptr<slashdot>(new slashdot(out,alarm_s,refresh));
#endif
    
#ifndef USE_OLD_SIGNAL
    struct sigaction handler;
    sigemptyset(&handler.sa_mask);
    handler.sa_handler=slashdot::handler(&(*sdconn));

#ifdef SA_RESTART
    handler.sa_flags=SA_RESTART;
#else
    handler.sa_flags=0;
#endif // SA_RESTART

    if(sigaction(SIGALRM,&handler,NULL)<0)
      throw error(strerror(errno));
#else
    if(signal(SIGALRM,slashdot::handler(&(*sdconn)))==SIG_ERR)
      throw error(strerror(errno));
#endif // USE_OLD_SIGNAL

#ifndef USE_OLD_TIMER
    itimerval timer;
    memset(static_cast<void*>(&timer),0,sizeof(itimerval));
    timer.it_interval.tv_sec=timer.it_value.tv_sec=alarm_s;
    if(setitimer(ITIMER_REAL,&timer,NULL)==-1)
      throw error(strerror(errno));
#else
    alarm(static_cast<unsigned int>(alarm_s));
#endif

    bool pause=false;
    for(;;)
    {
      XEvent event;
      XNextEvent(display,&event);
      switch(event.type)
      {
      case Expose:
	if(event.xexpose.count>0)
	  break;
	sdconn->reprint_article();
	break;
      case KeyPress:
	{
	  char buffer[BUFFER_SIZE];
	  KeySym keysym=0;
	  XLookupString(reinterpret_cast<XKeyEvent*>(&event),buffer,BUFFER_SIZE
			,&keysym,NULL);
	  switch(keysym)
	  {
	  case XK_Pause:
	    if(!pause)
	    {
#ifndef USE_OLD_TIMER
	      timer.it_value.tv_sec=0;
	      timer.it_interval.tv_sec=0;
	      if(setitimer(ITIMER_REAL,&timer,NULL)==-1)
		throw error(strerror(errno));
#else
	      alarm(0);
#endif
	    }
	    else
	    {
#ifndef USE_OLD_TIMER
	      timer.it_interval.tv_sec=timer.it_value.tv_sec=alarm_s;
	      if(setitimer(ITIMER_REAL,&timer,NULL)==-1)
		throw error(strerror(errno));
#else
	      alarm(static_cast<unsigned int>(alarm_s));
#endif
	    }
	  case XK_Up:
	    sdconn->print_prev_article();
	  case XK_Down:
	    sdconn->print_next_article();
	    break;
	  default:
	    std::cerr << "undefined key: " << keysym << std::endl;
	  }
	}
	break;
      case ButtonPress:
	switch(reinterpret_cast<XButtonEvent*>(&event)->button)
	{
	case Button1: //left
	  sdconn->openurl(browser.c_str());
	  break;
	case Button2:
	case Button5: //scroll down
	  sdconn->print_next_article();
	  break;
	case Button3:
	case Button4: //scroll up
	  sdconn->print_prev_article();
	  break;
	}
	break;
      }
    }
    XCloseDisplay(display);
  }
  catch(error &err)
  {
    std::cerr << "error: " << err.description() << std::endl;
    return 1;
  }
  catch(...)
  {
    std::cerr << "unkown error" << std::endl;
    return 1;
  }
}
